#!/usr/bin/env python

"""Check if data files are here and which version they have. Contact the
   server for the latest version and update them if they are outdated.
"""

import sys
import os
import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.DEBUG,
                    stream=sys.stdout)
import yaml
import hashlib
import urllib
# my ones
import hwrt.utils as utils


def is_file_consistent(local_path_file, md5_hash):
    """Check if file is there and if the md5_hash is correct."""
    if not os.path.isfile(local_path_file):
        return False
    if hashlib.md5(open(local_path_file, 'rb').read()).hexdigest() != md5_hash:
        return False
    return True


def get_parser():
    from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
    parser = ArgumentParser(description=__doc__,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    return parser


def main():
    # Read config file. This has to get updated via git
    PROJECT_ROOT = utils.get_project_root()
    infofile = os.path.join(PROJECT_ROOT, "raw-datasets/info.yml")
    logging.info("Read '%s'...", infofile)
    with open(infofile, 'r') as ymlfile:
        datasets = yaml.load(ymlfile)
    for dataset in datasets:
        local_path_file = os.path.join(PROJECT_ROOT, dataset['online_path'])
        i = 0
        while not is_file_consistent(local_path_file, dataset['md5']) and i < 3:
            if os.path.isfile(local_path_file):
                logging.info("MD5 codes differ. ")
                logging.info("The file size of the downloaded file is %s." %
                             utils.sizeof_fmt(os.path.getsize(local_path_file))
                             )
            logging.info("Download the file '%s'...", dataset['online_path'])
            urllib.urlretrieve(dataset['url'], local_path_file)
            i += 1
        if i < 10:
            logging.info("Found '%s'.", dataset['online_path'])


if __name__ == "__main__":
    get_parser().parse_args()
    main()



#!/usr/bin/env python

import os
import logging
import sys
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.DEBUG,
                    stream=sys.stdout)
import subprocess
import time
import re
import yaml
# mine
import hwrt.utils as utils


def get_error_from_logfile(logfile):
    with open(logfile) as f:
        log_content = f.read()
    pattern = re.compile("errors = (\d\.\d+)")
    errors = pattern.findall(log_content)
    # Get number in logfile
    _, ianhang = logfile.split("_")
    i, _ = ianhang.split(".")
    i = int(i)
    # Back up in case of error with that file
    while len(errors) == 0:
        logfile = "testresult_%03d.txt" % (i-1)
        with open(logfile) as f:
            log_content = f.read()
        pattern = re.compile("errors = (\d\.\d+)")
        errors = pattern.findall(log_content)
        print("Error with %i. Skip to %i." % (i, i-1))
        i -= 1
    error = float(errors[-1])
    return error


def test_model(model_folder, basename, test_file):
    model_src = utils.get_latest_working_model(model_folder)
    if model_src is None:
        logging.error("No model with basename '%s' found in '%s'.",
                      basename,
                      model_folder)
    else:
        os.chdir(model_folder)
        time_prefix = time.strftime("%Y-%m-%d-%H-%M")
        logging.info("Evaluate '%s'...", model_src)
        logfile = "%s-testing.log" % time_prefix
        with open(logfile, "w") as log, open(model_src, "r") as model_src_p:
            p = subprocess.Popen(['nntoolkit', 'test', '--batch-size', '1',
                                  test_file],
                                 stdin=model_src_p,
                                 stderr=log)
            ret = p.wait()
            if ret != 0:
                logging.error("nntoolkit finished with ret code %s", str(ret))
                sys.exit()

        return get_error_from_logfile(logfile)


def main(model_folder, run_native=False):
    PROJECT_ROOT = utils.get_project_root()
    # Get model description
    model_description_file = os.path.join(model_folder, "info.yml")
    # Read the model description file
    with open(model_description_file, 'r') as ymlfile:
        model_description = yaml.load(ymlfile)
    test_data_path = os.path.join(PROJECT_ROOT,
                                  model_description["data-source"],
                                  "testdata.pfile")
    error = test_model(model_folder,
                       "model",
                       test_data_path)
    if run_native:
        logging.info("Error: %0.4f", error)
    return error


def get_parser():
    PROJECT_ROOT = utils.get_project_root()

    # Get latest model folder
    models_folder = os.path.join(PROJECT_ROOT, "models")
    latest_model = utils.get_latest_folder(models_folder)

    # Get command line arguments
    from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
    parser = ArgumentParser(description=__doc__,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-m", "--model",
                        dest="model",
                        help="where is the model folder (with a info.yml)?",
                        metavar="FOLDER",
                        type=lambda x: utils.is_valid_folder(parser, x),
                        default=latest_model)
    return parser


if __name__ == "__main__":
    args = get_parser().parse_args()
    main(args.model, run_native=True)



#!/usr/bin/env python

"""Create preprocessed dataset."""

from __future__ import print_function
import logging
import sys
import os
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.DEBUG,
                    stream=sys.stdout)
import cPickle as pickle
import time
import yaml
# mine
import utils
import preprocessing
import HandwrittenData

sys.modules['hwrt.HandwrittenData'] = HandwrittenData
sys.modules['HandwrittenData'] = HandwrittenData


def main(folder):
    raw_datapath, outputpath, p_queue = get_parameters(folder)
    create_preprocessed_dataset(raw_datapath, outputpath, p_queue)
    utils.create_run_logfile(folder)


def get_parameters(folder):
    PROJECT_ROOT = utils.get_project_root()
    # Read the model description file
    with open(os.path.join(folder, "info.yml"), 'r') as ymlfile:
        preprocessing_description = yaml.load(ymlfile)

    # Get the path of the raw data
    raw_datapath = os.path.join(PROJECT_ROOT,
                                preprocessing_description['data-source'])
    # Get the path were the preprocessed file should be put
    outputpath = os.path.join(folder, "data.pickle")

    # Get the preprocessing queue
    tmp = preprocessing_description['queue']
    preprocessing_queue = preprocessing.get_preprocessing_queue(tmp)
    return (raw_datapath, outputpath, preprocessing_queue)


def create_preprocessed_dataset(path_to_data, outputpath, preprocessing_queue):
    # Log everything
    logging.info("Data soure %s" % path_to_data)
    logging.info("Output will be stored in %s" % outputpath)
    tmp = "Preprocessing Queue:\n"
    for el in preprocessing_queue:
        tmp += str(el) + "\n"
    logging.info(tmp)
    # Load from pickled file
    if not os.path.isfile(path_to_data):
        logging.info(("'%s' does not exist. Please either abort this script "
                      "or update the data location."), path_to_data)
        raw_dataset_path = utils.choose_raw_dataset()
        # Get project-relative path
        raw_dataset_path = "raw-datasets" + \
                           raw_dataset_path.split("raw-datasets")[1]
        print(raw_dataset_path)
        sys.exit()  # TODO: Update model!
    logging.info("Start loading data...")
    loaded = pickle.load(open(path_to_data))
    raw_datasets = loaded['handwriting_datasets']
    logging.info("Start applying preprocessing methods")
    start_time = time.time()
    for i, raw_dataset in enumerate(raw_datasets):
        if i % 10 == 0 and i > 0:
            utils.print_status(len(raw_datasets), i, start_time)
        # Do the work
        raw_dataset['handwriting'].preprocessing(preprocessing_queue)
    sys.stdout.write("\r%0.2f%% (done)\033[K\n" % (100))
    print("")
    pickle.dump({'handwriting_datasets': raw_datasets,
                 'formula_id2latex': loaded['formula_id2latex'],
                 'preprocessing_queue': preprocessing_queue
                 },
                open(outputpath, "wb"),
                2)


if __name__ == '__main__':
    PROJECT_ROOT = utils.get_project_root()

    # Get latest model description file
    preprocessed_folder = os.path.join(PROJECT_ROOT, "preprocessed")
    latest_preprocessed = utils.get_latest_folder(preprocessed_folder)

    # Get command line arguments
    from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
    parser = ArgumentParser(description=__doc__,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-f", "--folder",
                        dest="folder",
                        help="where is the preprocessing folder "
                             "(that contains a info.yml)?",
                        metavar="FOLDER",
                        type=lambda x: utils.is_valid_folder(parser, x),
                        default=latest_preprocessed)
    args = parser.parse_args()
    main(args.folder)



#!/usr/bin/env python
"""
Display a raw_data_id.
"""

import os
import yaml
import MySQLdb
import MySQLdb.cursors
# mine
import hwrt
from hwrt.HandwrittenData import HandwrittenData
import hwrt.utils as utils
import hwrt.preprocessing as preprocessing
import hwrt.features as features
import hwrt.data_multiplication as data_multiplication


def fetch_data(raw_data_id):
    """Get the data from raw_data_id from the server."""
    # Import configuration file
    cfg = utils.get_database_configuration()

    # Establish database connection
    connection = MySQLdb.connect(host=cfg[args.mysql]['host'],
                                 user=cfg[args.mysql]['user'],
                                 passwd=cfg[args.mysql]['passwd'],
                                 db=cfg[args.mysql]['db'],
                                 cursorclass=MySQLdb.cursors.DictCursor)
    cursor = connection.cursor()

    # Download dataset
    sql = ("SELECT `id`, `data` "
           "FROM `wm_raw_draw_data` WHERE `id`=%i") % args.id
    cursor.execute(sql)
    data = cursor.fetchone()
    return data


def display_data(raw_data_string, raw_data_id, model_folder):
    PROJECT_ROOT = utils.get_project_root()

    # Get model description
    model_description_file = os.path.join(model_folder, "info.yml")
    with open(model_description_file, 'r') as ymlfile:
        model_description = yaml.load(ymlfile)

    # Get the feature description
    feature_description_file = os.path.join(
        PROJECT_ROOT,
        model_description['data-source'],
        "info.yml")
    with open(feature_description_file, 'r') as ymlfile:
        feature_description = yaml.load(ymlfile)

    # Get the preprocessing description
    preprocessing_description_file = os.path.join(
        PROJECT_ROOT,
        feature_description['data-source'],
        "info.yml")
    with open(preprocessing_description_file, 'r') as ymlfile:
        preprocessing_description = yaml.load(ymlfile)

    # Get Handwriting
    a = HandwrittenData(raw_data_string, raw_data_id=raw_data_id)

    # Get the preprocessing queue
    tmp = preprocessing_description['queue']
    preprocessing_queue = preprocessing.get_preprocessing_queue(tmp)
    a.preprocessing(preprocessing_queue)

    tmp = feature_description['features']
    feature_list = features.get_features(tmp)
    x = a.feature_extraction(feature_list)
    t = [round(el, 3) for el in x]
    print("Features:")
    print(t)

    # Get the list of data multiplication algorithms
    mult_queue = data_multiplication.get_data_multiplication_queue(
        feature_description['data-multiplication'])

    # Multiply traing_set
    training_set = [a]
    for algorithm in mult_queue:
        new_trning_set = []
        for recording in training_set:
            samples = algorithm(recording)
            for sample in samples:
                new_trning_set.append(sample)
        training_set = new_trning_set

    # Display it
    for recording in training_set:
        recording.show()


def get_parser():
    PROJECT_ROOT = utils.get_project_root()

    # Get latest model description file
    models_folder = os.path.join(PROJECT_ROOT, "models")
    latest_model = utils.get_latest_in_folder(models_folder, ".yml")

    # Parse command line arguments
    from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
    parser = ArgumentParser(description=__doc__,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-i", "--id", dest="id", default=279062,
                        type=int,
                        help="which RAW_DATA_ID do you want?")
    parser.add_argument("--mysql", dest="mysql", default='mysql_online',
                        help="which mysql configuration should be used?")
    parser.add_argument("-m", "--model",
                        dest="model",
                        help="where is the model folder (with a info.yml)?",
                        metavar="FOLDER",
                        type=lambda x: utils.is_valid_folder(parser, x),
                        default=latest_model)
    return parser

if __name__ == '__main__':
    args = get_parser().parse_args()

    # do something
    data = fetch_data(args.id)
    if data is None:
        print("RAW_DATA_ID %i does not exist." % args.id)
    else:
        print("hwrt version: %s" % hwrt.__version__)
        display_data(data['data'], data['id'], args.model)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Preprocessing algorithms.

Each algorithm works on the HandwrittenData class. They have to be applied like
this:

 >>> a = HandwrittenData(...)
 >>> preprocessing_queue = [Scale_and_shift(), \
                           Stroke_connect(), \
                           Douglas_peucker(EPSILON=0.2), \
                           Space_evenly(number=100)]
 >>> a.preprocessing(preprocessing_queue)
"""

import numpy
import inspect
from scipy.interpolate import interp1d
import math
import logging
import sys
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.DEBUG,
                    stream=sys.stdout)
# mine
import HandwrittenData


def _euclidean_distance(p1, p2):
    """Calculate the euclidean distance of two 2D points."""
    return math.sqrt((p1["x"]-p2["x"])**2 + (p1["y"]-p2["y"])**2)


def _flatten(two_dimensional_list):
    """Flatten a 2D list into a 1D list"""
    return [i for inner_list in two_dimensional_list for i in inner_list]


def get_class(name):
    """Get function pointer by string."""
    clsmembers = inspect.getmembers(sys.modules[__name__], inspect.isclass)
    for string_name, act_class in clsmembers:
        if string_name == name:
            return act_class
    logging.debug("Unknown class '%s'.", name)
    return None


def get_preprocessing_queue(model_description_preprocessing):
    """Get preprocessing queue from a list of dictionaries

    >>> l = [{'Remove_duplicate_time': None}, \
             {'Scale_and_shift': [{'center': True}]} \
            ]
    >>> get_preprocessing_queue(l)
    [Remove_duplicate_time, Scale_and_shift
     - center: True
     - max_width: 1
     - max_height: 1
    ]
    """
    preprocessing_queue = []
    for preprocessing in model_description_preprocessing:
        for alg, params in preprocessing.items():
            alg = get_class(alg)
            if params is None:
                preprocessing_queue.append(alg())
            else:
                parameters = {}
                for dicts in params:
                    for param_name, param_value in dicts.items():
                        parameters[param_name] = param_value
                preprocessing_queue.append(alg(**parameters))
    return preprocessing_queue

# Only preprocessing classes follow
# Everyone must have a __str__, __repr__ and __call__
# where
# * __call__ must take exactly one argument of type HandwrittenData
# * __call__ must call the Handwriting.set_points


class Remove_duplicate_time(object):
    """If a recording has two points with the same timestamp, than the second
       one will be discarded."""
    def __repr__(self):
        return "Remove_duplicate_time"

    def __str__(self):
        return "remove duplicate time"

    def __call__(self, handwritten_data):
        assert isinstance(handwritten_data, HandwrittenData.HandwrittenData), \
            "handwritten data is not of type HandwrittenData, but of %r" % \
            type(handwritten_data)
        pointlist = handwritten_data.get_pointlist()
        new_pointlist = []
        t = []
        for stroke in pointlist:
            new_stroke = []
            for point in stroke:
                if point['time'] not in t:
                    new_stroke.append(point)
                    t.append(point['time'])
            new_pointlist.append(new_stroke)
        handwritten_data.set_pointlist(new_pointlist)


class Remove_points(object):
    """Remove all single point strokes from the recording, except if
       the whole recording consists of points only.
    """
    def __repr__(self):
        return "Remove_points"

    def __str__(self):
        return "remove points"

    def __call__(self, handwritten_data):
        assert isinstance(handwritten_data, HandwrittenData.HandwrittenData), \
            "handwritten data is not of type HandwrittenData, but of %r" % \
            type(handwritten_data)
        pointlist = handwritten_data.get_pointlist()
        has_nonpoint_stroke = False
        # Check if recording has non-point stroke:
        for stroke in pointlist:
            if len(stroke) > 1:
                has_nonpoint_stroke = True
        if has_nonpoint_stroke:
            new_pointlist = []
            for stroke in pointlist:
                if len(stroke) > 1:
                    new_pointlist.append(stroke)
            handwritten_data.set_pointlist(new_pointlist)


class Scale_and_shift(object):
    """ Scale a recording so that it fits into a unit square. This keeps the
        aspect ratio. Then the recording is shifted. The default way is to
        shift it so that the recording is in [0, 1] × [0,1]. However, it
        can also be used to be centered within [-1, 1] × [-1, 1] around the
        origin (0, 0) by setting center=True and center_other=True.
    """
    def __init__(self, center=False, max_width=1., max_height=1.,
                 width_add=0, height_add=0, center_other=False):
        self.center = center
        self.max_width = max_width
        self.max_height = max_height
        self.width_add = width_add
        self.height_add = height_add
        self.center_other = center_other

    def __repr__(self):
        return ("Scale_and_shift\n"
                " - center: %r\n"
                " - max_width: %i\n"
                " - max_height: %i\n") % \
            (self.center, self.max_width, self.max_height)

    def __str__(self):
        return ("Scale and shift\n"
                " - center: %r\n"
                " - max_width: %i\n"
                " - max_height: %i\n") % \
            (self.center, self.max_width, self.max_height)

    def get_scale_and_shift_parameters(self, handwritten_data):
        """ Take a list of points and calculate the factors for scaling and
            moving it so that it's in the unit square. Keept the aspect
            ratio.
            Optionally center the points inside of the unit square.
        """
        a = handwritten_data.get_bounding_box()

        width = a['maxx'] - a['minx'] + self.width_add
        height = a['maxy'] - a['miny'] + self.height_add

        factorX, factorY = 1, 1
        if width != 0:
            factorX = self.max_width/width

        if height != 0:
            factorY = self.max_height/height

        factor = min(factorX, factorY)
        addx, addy = 0, 0

        if self.center:
            # Only one dimension (x or y) has to be centered (the smaller one)
            add = -(factor/(2*max(factorX, factorY)))

            if factor == factorX:
                addy = add
                if self.center_other:
                    addx = -(width*factor/2.0)
            else:
                addx = add
                if self.center_other:
                    addy = -(height*factor/2.0)

        return {"factor": factor, "addx": addx, "addy": addy,
                "minx": a['minx'], "miny": a['miny'], "mint": a['mint']}

    def __call__(self, handwritten_data):
        assert isinstance(handwritten_data, HandwrittenData.HandwrittenData), \
            "handwritten data is not of type HandwrittenData, but of %r" % \
            type(handwritten_data)

        tmp = self.get_scale_and_shift_parameters(handwritten_data)
        factor, addx, addy = tmp['factor'], tmp['addx'], tmp['addy']
        minx, miny, mint = tmp['minx'], tmp['miny'], tmp['mint']

        pointlist = handwritten_data.get_pointlist()
        for strokenr, stroke in enumerate(pointlist):
            for key, p in enumerate(stroke):
                pointlist[strokenr][key] = {
                    "x": (p["x"] - minx) * factor + addx,
                    "y": (p["y"] - miny) * factor + addy,
                    "time": p["time"] - mint}
                if "pen_down" in p:
                    pointlist[strokenr][key]["pen_down"] = p["pen_down"]
        handwritten_data.set_pointlist(pointlist)
        assert self.max_width - handwritten_data.get_width() >= -0.00001, \
            "max_width: %0.5f; width: %0.5f" % (self.max_width,
                                                handwritten_data.get_width())
        assert self.max_height - handwritten_data.get_height() >= -0.00001, \
            "max_height: %0.5f; height: %0.5f" % \
            (self.max_height, handwritten_data.get_height())


class Space_evenly(object):
    """Space the points evenly in time over the complete recording. The
       parameter 'number' defines how many"""
    def __init__(self, number=100, kind='cubic'):
        self.number = number
        self.kind = kind

    def __repr__(self):
        return ("Space_evenly\n"
                " - number: %i\n"
                " - kind: %s\n") % \
            (self.number, self.kind)

    def __str__(self):
        return ("Space evenly\n"
                " - number: %i\n"
                " - kind: %s\n") % \
            (self.number, self.kind)

    def __call__(self, handwritten_data):
        assert isinstance(handwritten_data, HandwrittenData.HandwrittenData), \
            "handwritten data is not of type HandwrittenData, but of %r" % \
            type(handwritten_data)
        # Make sure that the lists are sorted
        pointlist = handwritten_data.get_sorted_pointlist()

        for i in range(len(pointlist)-1):
            # The last point of the previous stroke should be lower than the
            # first point of the next stroke
            assert (pointlist[i][-1]["time"] <= pointlist[i+1][0]["time"]), \
                ("Something is wrong with the time. The last point of "
                 "stroke %i has a time of %0.2f, but the first point of "
                 "stroke %i has a time of %0.2f. See raw_data_id %s") % \
                (i,
                 pointlist[i][-1]["time"],
                 i+1,
                 pointlist[i+1][0]["time"],
                 str(handwritten_data.raw_data_id))

        # calculate "pen_down" strokes
        times = []
        for i, stroke in enumerate(pointlist):
            stroke_info = {"start": stroke[0]['time'],
                           "end": stroke[-1]['time'],
                           "pen_down": True}
            # set up variables for interpolation
            x, y, t = [], [], []
            for point in stroke:
                if point['time'] not in t:
                    x.append(point['x'])
                    y.append(point['y'])
                    t.append(point['time'])
            x, y = numpy.array(x), numpy.array(y)
            if len(t) == 1:
                # constant interpolation
                fx, fy = lambda x: float(x), lambda y: float(y)
            elif len(t) == 2:
                # linear interpolation
                fx, fy = interp1d(t, x, 'linear'), interp1d(t, y, 'linear')
            elif len(t) == 3:
                # quadratic interpolation
                fx = interp1d(t, x, 'quadratic')
                fy = interp1d(t, y, 'quadratic')
            else:
                fx, fy = interp1d(t, x, self.kind), interp1d(t, y, self.kind)
            stroke_info['fx'] = fx
            stroke_info['fy'] = fy
            times.append(stroke_info)

        # Model "pen_up" strokes
        for i in range(len(pointlist) - 1):
            stroke_info = {"start": pointlist[i][-1],
                           "end": pointlist[i+1][0],
                           "pen_down": False}
            x, y, t = [], [], []
            for point in [pointlist[i][-1], pointlist[i+1][0]]:
                if point['time'] not in t:
                    x.append(point['x'])
                    y.append(point['y'])
                    t.append(point['time'])
                else:
                    # TODO: That should not happen
                    pass
            if len(x) == 1:
                # constant interpolation
                fx, fy = lambda x: float(x), lambda y: float(y)
            else:
                # linear interpolation
                x, y = numpy.array(x), numpy.array(y)
                fx = interp1d(t, x, kind='linear')
                fy = interp1d(t, y, kind='linear')
            stroke_info['fx'] = fx
            stroke_info['fy'] = fy
            times.append(stroke_info)

        new_pointlist = []

        tnew = numpy.linspace(pointlist[0][0]['time'],
                              pointlist[-1][-1]['time'],
                              self.number)

        for time in tnew:
            for stroke_intervall in times:
                if stroke_intervall["start"] <= time <= stroke_intervall["end"]:
                    x = float(stroke_intervall['fx'](time))
                    y = float(stroke_intervall['fy'](time))
                    time = float(time)
                    new_pointlist.append({'x': x, 'y': y, 'time': time,
                                          'pen_down':
                                          stroke_intervall['pen_down']})
        handwritten_data.set_pointlist([new_pointlist])


class Space_evenly_per_stroke(object):
    """Space the points evenly for every single stroke seperatly. """
    def __init__(self, number=100, kind='cubic'):
        self.number = number
        self.kind = kind

    def __repr__(self):
        return ("Space_evenly_per_stroke\n"
                " - number: %i\n"
                " - kind: %s\n") % \
            (self.number, self.kind)

    def __str__(self):
        return ("Space evenly per stroke\n"
                " - number: %i\n"
                " - kind: %s\n") % \
            (self.number, self.kind)

    def __call__(self, handwritten_data):
        assert isinstance(handwritten_data, HandwrittenData.HandwrittenData), \
            "handwritten data is not of type HandwrittenData, but of %r" % \
            type(handwritten_data)
        pointlist = handwritten_data.get_pointlist()
        new_pointlist = []

        for stroke in pointlist:
            new_stroke = []
            if len(stroke) < 4:
                # Don't do anything if there are less than 4 points
                new_stroke = stroke
            else:
                stroke = sorted(stroke, key=lambda p: p['time'])

                x, y, t = [], [], []

                for point in stroke:
                    x.append(point['x'])
                    y.append(point['y'])
                    t.append(point['time'])

                x, y = numpy.array(x), numpy.array(y)
                failed = False
                try:
                    fx = interp1d(t, x, kind=self.kind)
                    fy = interp1d(t, y, kind=self.kind)
                except Exception as e:
                    if handwritten_data.raw_data_id is not None:
                        logging.debug("spline failed for raw_data_id %i",
                                      handwritten_data.raw_data_id)
                    else:
                        logging.debug("spline failed")
                    logging.debug(e)
                    failed = True

                tnew = numpy.linspace(t[0], t[-1], self.number)

                # linear interpolation fallback due to
                # https://github.com/scipy/scipy/issues/3868
                if failed:
                    try:
                        fx = interp1d(t, x, kind='linear')
                        fy = interp1d(t, y, kind='linear')
                        failed = False
                    except Exception as e:
                        raise e

                for x, y, t in zip(fx(tnew), fy(tnew), tnew):
                    new_stroke.append({'x': x, 'y': y, 'time': t})
            new_pointlist.append(new_stroke)
        handwritten_data.set_pointlist(new_pointlist)


class Douglas_peucker(object):
    """
     Apply the Douglas-Peucker algorithm to each stroke of $pointlist
     seperately.
    """
    def __init__(self, EPSILON):
        self.EPSILON = EPSILON

    def __repr__(self):
        return "Douglas_peucker (EPSILON: %0.2f)\n" % self.EPSILON

    def __str__(self):
        return "Douglas_peucker (EPSILON: %0.2f)\n" % self.EPSILON

    def DouglasPeucker(self, PointList, EPSILON):
        def LotrechterAbstand(p3, p1, p2):
            """
            Calculate the distance from p3 to the stroke defined by p1 and p2.
            :param p1: start of stroke
            :type p1: dictionary with "x" and "y"
            :param p2: end of stroke
            :type p2: dictionary with "x" and "y"
            :param p3: point
            :type p3: dictionary with "x" and "y"
            """
            x3 = p3['x']
            y3 = p3['y']

            px = p2['x']-p1['x']
            py = p2['y']-p1['y']

            something = px*px + py*py
            if (something == 0):
                # TODO: really?
                return 0

            u = ((x3 - p1['x']) * px + (y3 - p1['y']) * py) / something

            if u > 1:
                u = 1
            elif u < 0:
                u = 0

            x = p1['x'] + u * px
            y = p1['y'] + u * py

            dx = x - x3
            dy = y - y3

            # Note: If the actual distance does not matter,
            # if you only want to compare what this function
            # returns to other results of this function, you
            # can just return the squared distance instead
            # (i.e. remove the sqrt) to gain a little performance

            dist = math.sqrt(dx*dx + dy*dy)
            return dist

        # Finde den Punkt mit dem größten Abstand
        dmax = 0
        index = 0
        for i in range(1, len(PointList)):
            d = LotrechterAbstand(PointList[i], PointList[0], PointList[-1])
            if d > dmax:
                index = i
                dmax = d

        # Wenn die maximale Entfernung größer als EPSILON ist, dann rekursiv
        # vereinfachen
        if dmax >= EPSILON:
            # Recursive call
            recResults1 = self.DouglasPeucker(PointList[0:index], EPSILON)
            recResults2 = self.DouglasPeucker(PointList[index:], EPSILON)

            # Ergebnisliste aufbauen
            ResultList = recResults1[:-1] + recResults2
        else:
            ResultList = [PointList[0], PointList[-1]]

        # Ergebnis zurückgeben
        return ResultList

    def __call__(self, handwritten_data):
        assert isinstance(handwritten_data, HandwrittenData.HandwrittenData), \
            "handwritten data is not of type HandwrittenData, but of %r" % \
            type(handwritten_data)
        pointlist = handwritten_data.get_pointlist()

        for i in range(0, len(pointlist)):
            pointlist[i] = self.DouglasPeucker(pointlist[i], self.EPSILON)
        handwritten_data.set_pointlist(pointlist)


class Stroke_connect(object):
    """Detect if strokes were probably accidentially disconnected. If that is
       the case, connect them.
    """
    def __init__(self, minimum_distance=0.05):
        self.minimum_distance = minimum_distance

    def __repr__(self):
        return "Stroke_connect (minimum_distance: %0.2f)" % \
            self.minimum_distance

    def __str__(self):
        return "Stroke connect (minimum_distance: %0.2f)" % \
            self.minimum_distance

    def __call__(self, handwritten_data):
        assert isinstance(handwritten_data, HandwrittenData.HandwrittenData), \
            "handwritten data is not of type HandwrittenData, but of %r" % \
            type(handwritten_data)
        pointlist = handwritten_data.get_pointlist()

        # Connecting strokes makes only sense when there are multiple strokes
        if len(pointlist) > 1:
            strokes = []
            last_appended = False
            i = 0
            while i < len(pointlist)-1:
                last_point = pointlist[i][-1]
                first_point = pointlist[i+1][0]
                if _euclidean_distance(last_point, first_point) < \
                   self.minimum_distance:
                    strokes.append(pointlist[i]+pointlist[i+1])
                    pointlist[i+1] = strokes[-1]
                    if i == len(pointlist)-2:
                        last_appended = True
                    i += 1
                else:
                    strokes.append(pointlist[i])
                i += 1
            if not last_appended:
                strokes.append(pointlist[-1])
            handwritten_data.set_pointlist(strokes)


class Dot_reduction(object):
    """Reduce strokes where the maximum distance between points is below a
       threshold to a single dot.
    """
    def __init__(self, threshold):
        self.threshold = threshold

    def __repr__(self):
        return "Dot_reduction (threshold: %0.2f)" % \
            self.threshold

    def __str__(self):
        return "Dot_reduction (threshold: %0.2f)" % \
            self.threshold

    def __call__(self, handwritten_data):
        assert isinstance(handwritten_data, HandwrittenData.HandwrittenData), \
            "handwritten data is not of type HandwrittenData, but of %r" % \
            type(handwritten_data)

        def get_max_distance(L):
            """Find the maximum distance between two points in a list of points
            :param L: list of points
            :type L: list
            :returns: maximum distance bewtween two points
            :rtype: float
            """
            if len(L) <= 1:
                return -1
            else:
                max_dist = _euclidean_distance(L[0], L[1])
                for i in range(len(L)-1):
                    for j in range(i+1, len(L)):
                        max_dist = max(_euclidean_distance(L[i], L[j]),
                                       max_dist)
                return max_dist

        def get_average_point(L):
            """Calculate the average point.
            :param L: list of points
            :type L: list
            :returns: a single point
            :rtype: dict
            """
            x, y, t = 0, 0, 0
            for point in L:
                x += point['x']
                y += point['y']
                t += point['time']
            x = float(x) / len(L)
            y = float(y) / len(L)
            t = float(t) / len(L)
            return {'x': x, 'y': y, 'time': t}

        new_pointlist = []
        pointlist = handwritten_data.get_pointlist()
        for stroke in pointlist:
            new_stroke = stroke
            if len(stroke) > 1 and get_max_distance(stroke) < self.threshold:
                new_stroke = [get_average_point(stroke)]
            new_pointlist.append(new_stroke)

        handwritten_data.set_pointlist(new_pointlist)


class Wild_point_filter(object):
    """Find wild points and remove them. The threshold means
       speed in pixels / ms.
    """
    def __init__(self, threshold):
        """The threshold is a speed threshold"""
        self.threshold = threshold

    def __repr__(self):
        return "Wild_point_filter"

    def __str__(self):
        return "Wild point filter (threshold: %0.2f)" % \
            self.threshold

    def __call__(self, handwritten_data):
        assert isinstance(handwritten_data, HandwrittenData.HandwrittenData), \
            "handwritten data is not of type HandwrittenData, but of %r" % \
            type(handwritten_data)
        new_pointlist = []
        pointlist = handwritten_data.get_sorted_pointlist()
        for stroke in pointlist:
            new_stroke = []
            for last_point, point in zip(stroke, stroke[1:]):
                space_dist = math.hypot(last_point['x'] - point['x'],
                                        last_point['y'] - point['y'])
                time_dist = float(point['time'] - last_point['time'])
                if time_dist == 0:
                    continue
                speed = space_dist/time_dist
                if speed < self.threshold:
                    new_stroke.append(point)

            new_pointlist.append(new_stroke)
        # Bounding box criterion:
        # If the distance from point to all others strokes bounding boxes is
        # more than 1/5 of the whole size, it is a wild point


class Weighted_average_smoothing(object):
    """Smooth every stroke by a weighted average. """
    def __init__(self, theta):
        """Theta is a list of 3 non-negative numbers"""
        assert len(theta) == 3, \
            "theta has length %i, but should have length 3" % \
            len(theta)
        theta = map(float, theta)
        # Normalize parameters to a sum of 1
        self.theta = list(1./sum(theta) * numpy.array(theta))

    def __repr__(self):
        return "Weighted_average_smoothing"

    def __str__(self):
        return "Weighted average smoothing (theta: %s)" % \
            self.theta

    def calculate_average(self, points):
        assert len(self.theta) == len(points), \
            "points has length %i, but should have length %i" % \
            (len(points), len(self.theta))
        new_point = {'x': 0, 'y': 0, 'time': 0}
        for key in new_point:
            new_point[key] = self.theta[0] * points[0][key] + \
                self.theta[1] * points[1][key] + \
                self.theta[2] * points[2][key]
        return new_point

    def __call__(self, handwritten_data):
        assert isinstance(handwritten_data, HandwrittenData.HandwrittenData), \
            "handwritten data is not of type HandwrittenData, but of %r" % \
            type(handwritten_data)
        new_pointlist = []
        pointlist = handwritten_data.get_sorted_pointlist()
        for stroke in pointlist:
            tmp = [stroke[0]]
            new_pointlist.append(tmp)
            if len(stroke) > 1:
                for i in range(1, len(stroke)-1):
                    points = [stroke[i-1], stroke[i], stroke[i+1]]
                    p = self.calculate_average(points)
                    new_pointlist[-1].append(p)
                new_pointlist[-1].append(stroke[-1])
        handwritten_data.set_pointlist(new_pointlist)

if __name__ == '__main__':
    import doctest
    doctest.testmod()
